#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>

#ifndef PLATFORM_KIND
#define Posix_Platform  0 
#define Mingw_Platform  1
#define MsVisC_Platform 2

#ifdef __MINGW32__
#define PLATFORM_KIND Mingw_Platform
#elif defined(__WIN32) || defined(WIN32)
#define PLATFORM_KIND MsVisC_Platform
#else
#define PLATFORM_KIND Posix_Platform    /* Posix/Linux/Unix, Microsoft MinGW or MSVisC. */ 
#endif
#endif

#if (PLATFORM_KIND==Posix_Platform)
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdint.h>
#include <errno.h>
#define closesocket close
#define ioctlsocket ioctl
#define WSAEINTR EINTR
#define WSAEWOULDBLOCK EWOULDBLOCK
#define WSAECONNREFUSED ECONNREFUSED
#define WSAECONNABORTED ECONNABORTED
#define WSAEAGAIN EAGAIN
#define WSAEPROTO EPROTO
#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#define wsa_get_ready() {}
#else //Posix
#include <windows.h>
#include <winsock2.h>
#ifndef R_OK
#define R_OK 2
#endif
#define WSAEAGAIN WSATRY_AGAIN
#define WSAEPROTO WSAEPROTONOSUPPORT
typedef long in_addr_t;
int wsa_is_ready = 0;
#define wsa_get_ready() if( !wsa_is_ready ) {WSADATA wd; WSAStartup( MAKEWORD(2,2), &wd );}
#endif // Posix

#define SOCKADDR_IN(s) ((struct sockaddr_in *)(s))
#define SOCKADDR_SA(s) ((struct sockaddr *)(s))

#ifndef __USE_MISC
/* Old compatibility names for C types.  */
typedef unsigned long int ulong;
typedef unsigned short int ushort;
typedef unsigned int uint;

#ifndef __USE_POSIX
#define strtok_r(s, sep, nextp) c_strtok_r(s, sep, nextp)
#endif
#endif

#ifndef c_new
#define c_new(t,n) ((t *)malloc( sizeof(t)*(n) ))
#define c_new0(t,n) ((t *)calloc( (n), sizeof(t) ))
#define c_dup(p,t,n) ((t *)memcpy( malloc( sizeof(t)*(n) ), (char *)(p), sizeof(t)*(n) ))
#define c_strdup(s) strcpy( malloc( strlen(s) ), (s) )
#endif

int recv_soc( int sockfd, char *line, int len );
/* create a socket on which I can listen and accept connections */
int net_server_new( unsigned short port );
/* returns a new socket from listening socket if successful, 0 otherwise */
int net_server_accept_connection( int server, char *remote, uint remotelen,
				  ushort *remoteport );
/* returns socket if successful, 0 otherwise */
int net_connection_new( const char *hostname, unsigned short port );
int net_connection_new_by_ip( unsigned int ip, unsigned short port );

typedef struct NetRecvBuf NetRecvBuf;
struct NetRecvBuf {
  int sockfd;
  char *buf;
  unsigned int len;
  NetRecvBuf *next;
} *net_recv_bufs = NULL;

/******************************************************************************/
/* create a socket on which I can listen and accept connections */
int net_server_new( unsigned short port ) {
  int sockfd;
  struct sockaddr_in sa;
  //int flags, socklen;
  const int on = 1;

  wsa_get_ready();

  /* Create socket */
  if( (sockfd = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
    perror( "net_create_server_socket: socket" );
    return( 0 ); 
  }

  /* set socket address parameters */
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = INADDR_ANY;  
  sa.sin_port = htons( port );

  /* Set REUSEADDR so we can reuse the port */
  if( setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(int) ) )
    perror( "net_create_server_socket: setsockopt(reuse)" );

  /* Bind */
  if( bind( sockfd, SOCKADDR_SA(&sa), sizeof(struct sockaddr_in) ) ) {
    perror( "net_create_server_socket: bind" );
    close( sockfd );
    return( 0 );
  }

#ifdef NET_DEAD_CODE
  /* Get the socket name */
  socklen = sizeof(struct sockaddr_in);
  if( getsockname( sockfd, SOCKADDR_SA(&sa), &socklen ) ) {
    perror( "net_create_server_socket: getsockname" );
    close( sockfd );
    return( 0 );
  }
#endif //NET_DEAD_CODE

  /* Listen */
  if( listen( sockfd, 10 ) != 0) {
    perror( "net_create_server_socket: listen" );
    close( sockfd );
    return( 0 );
  }
  
#if (PLATFORM_KIND==Posix_Platform)
#ifdef NET_MAKE_NONBLOCK
  // Make the socket non-blocking
  if( ioctlsocket( sockfd, FIONBIO, &on ) ) {
    perror( "net_create_server_socket: ioctl(FIONBIO)" );
    closesocket( sockfd );
    return( 0 );
  }
#endif //NET_MAKE_NONBLOCK

  /* set close_on_exec flag to TRUE */
  if( fcntl( sockfd, F_SETFD, 1 ) )
    perror( "net_create_server_socket: fcntl(SETFD)" );
#endif //UNIX

  return( sockfd );
}

/******************************************************************************/
/* create a socket on which I can listen and accept connections */
int createListeningSocketOnPort( unsigned short port ) {
  return( net_server_new( port ) );
}

/******************************************************************************/
/* returns a new socket no. from listening to socket s if successful, 0 otherwise */
int net_server_accept_connection( int server, char *remote, uint remotelen,
				  ushort *remoteport ) {
  struct sockaddr_storage sa;
#if (PLATFORM_KIND==Posix_Platform)
  socklen_t n;
#else
  int n;
#endif
  fd_set fdset;
  int client = 0, retry = 1, numfds = 0;

  wsa_get_ready();

  FD_ZERO( &fdset );
  FD_SET( server, &fdset );

  while( retry ) {
    retry = 0;
    numfds = select( server + 1, &fdset, NULL, NULL, NULL );
    if( numfds == SOCKET_ERROR ) { /* if there was an error waiting */
#if (PLATFORM_KIND!=Posix_Platform)
      errno = WSAGetLastError();
#endif
      if( errno == WSAEINTR ) {
	retry = 1;
      } else {
	perror( "net_accept_connections: select" );
	return( -1 );
      }
    } else {
      n = sizeof(struct sockaddr_storage);
      client = accept( server, SOCKADDR_SA(&sa), &n );
      if( client == INVALID_SOCKET ) {
#if (PLATFORM_KIND!=Posix_Platform)
	errno = WSAGetLastError();
#endif
	if( errno == WSAEWOULDBLOCK || errno == WSAECONNABORTED ||
	    errno == WSAEPROTO || errno == WSAEINTR ) {
	  retry = 1;
	} else {
	  perror( "net_accept_connections: accept" );
	  return( -1 );
	}
      }
    }
  }

  if( remote && remotelen ) {
    struct hostent *he;
    he = gethostbyaddr( (char *)&(SOCKADDR_IN(&sa)->sin_addr), 4, AF_INET );
    strncpy( remote, he->h_name, remotelen );
  }
  if( remoteport )
    *remoteport = ntohs( SOCKADDR_IN(&sa)->sin_port );

  return( client );
}

/******************************************************************************/
/* returns a new socket no. from listening to socket s if successful, 0 otherwise */
int acceptConnect( int sockfd ) {
  return( net_server_accept_connection( sockfd, NULL, 0, NULL ) );
}

/*************************************************************************************/
/* returns socket if successful, 0 otherwise       */
int net_connection_new( const char *hostname, unsigned short port ) {
  struct hostent *he;
  struct sockaddr_in sa;
  int sockfd, addrchar;
  const char *ptr;

  wsa_get_ready();

  addrchar = 0;
  for( ptr = hostname; (addrchar = (isdigit( *ptr ) || *ptr == '.')); ptr++ );
  
  if( addrchar ) {
    in_addr_t ip = inet_addr( hostname );
  
    if( ip != -1 )
      return( net_connection_new_by_ip( ip, port ) );

    fprintf( stderr, "net_connection_new: unknown host \"%s\"\n", hostname );
    return( 0 );
  }

  if( !(he = gethostbyname( hostname )) ) { 
    fprintf( stderr, "net_connection_new: unknown host \"%s\"\n", hostname );
    return( 0 );
  }
  memcpy( &sa.sin_addr, he->h_addr, he->h_length );

  sa.sin_port = htons( port );
  sa.sin_family = he->h_addrtype;
  while( 1 ) {
    if( (sockfd = socket( sa.sin_family, SOCK_STREAM, 0) ) == -1 ) {
      perror( "net_connection_new: socket" );
      close( sockfd );
      return( 0 ); 
    }

    if( !connect( sockfd, SOCKADDR_SA(&sa), sizeof(struct sockaddr_in) ) ) {
      return( sockfd );
    } else {
#if (PLATFORM_KIND!=Posix_Platform)
      errno = WSAGetLastError();
#endif
      if( errno == WSAECONNREFUSED || errno == WSAEINTR ) {
	close( sockfd );
      } else {
	perror( "net_connection_new: connect" ); 
	close( sockfd );
	return( 0 );
      }
    }
  }
}

int net_server_remote( int sockfd, char *buf, uint len ) {
  return 0;
}

/*************************************************************************************/
/* returns socket if successful, 0 otherwise       */
int makeConnect( const char *hostname, unsigned short port ) {
  return( net_connection_new( hostname, port ) );
}

/*******************************************************************************/
/* returns socket if successful, 0 otherwise       */
int net_connection_new_by_ip( unsigned int ip, unsigned short port ) {
  int sockfd;
  struct sockaddr_in sa;
 
  wsa_get_ready();

  while (1) {
    if( (sockfd = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
      perror( "net_connection_new_by_ip: socket" ); 
      close( sockfd );
      return( 0 ); 
    }

    memcpy( &sa.sin_addr, &ip, sizeof(in_addr_t) );
    sa.sin_port = htons( port );
    sa.sin_family = AF_INET;

    if( !connect( sockfd, SOCKADDR_SA(&sa), sizeof(struct sockaddr_in) ) ) {
      return( sockfd );
    } else {
#if (PLATFORM_KIND!=Posix_Platform)
      errno = WSAGetLastError();
#endif
      if( errno == WSAECONNREFUSED ) {
	perror( "net_connection_new_by_ip: connect" );
	close( sockfd );
      } else {
	perror( "net_connection_new_by_ip: connect" ); 
	close( sockfd );
	return 0;
      }
    }
  }
}

/*******************************************************************************/
/* returns socket if successful, 0 otherwise       */
int makeInetConnect( unsigned int inetaddr, unsigned short port ) {
  return( net_connection_new_by_ip( inetaddr, port ) );
}

/*******************************************************************************/
/* returns bytes sent if successful, -1 otherwise    */
int net_connection_send( int sockfd, char *line, int len ) {
  int bytes_sent = 0;
  fd_set fdset;
  int retry = 1;

  wsa_get_ready();

  while( retry ) {
    retry = 0;
    FD_ZERO( &fdset );
    FD_SET( sockfd, &fdset );

    if( select( sockfd + 1, NULL, &fdset, NULL, NULL) == SOCKET_ERROR ) {
#if (PLATFORM_KIND!=Posix_Platform)
      errno = WSAGetLastError();
#endif
      if( errno == WSAEINTR ) {
	retry = 1;
      } else {
	perror( "net_connection_send: select" );
	return( -1 );
      }
    } else {
      if( (bytes_sent = send( sockfd, line, len, 0 )) < 0 ) {
#if (PLATFORM_KIND!=Posix_Platform)
	errno = WSAGetLastError();
#endif
	if( errno == WSAEAGAIN || errno == WSAEINTR ) {
	  retry = 1;
	} else {
	  char ebuf[64];
	  sprintf( ebuf, "net_connection_send: send %d", errno );
	  perror( ebuf );
	  return( -1 );
	}
      } else if( !bytes_sent ) {
	perror( "net_connection_send: send going nowhere" );
	return( -1 );
      } else if( bytes_sent < len ) {
	line += bytes_sent;
	len -= bytes_sent;
	retry = 1;
      }
    }
  }

  return( bytes_sent );
}

int send_soc( int sockfd, char *line ) {
  // printf("WFSending: '%s'\n", line);
  fflush( stdout );
  return( net_connection_send( sockfd, line, strlen( line ) ) );
}

int send_hand_soc( int sockfd, char *line ) {
  int it;
  char *nl, handshake[2], size;
  
  while( line ) {
    if( (nl = strchr( line, '\n' )) )
      size = nl - line;
    else
      size = strlen( line );

    recv_soc( sockfd, handshake, 1 );
    it = net_connection_send( sockfd, &size, 1 );
    if( it != 1 )
      fprintf( stderr, "send_hand_soc: error: it!=1 but = %d\n", it );
     
    it = net_connection_send( sockfd, line, (int)size );
    if( it != (int)size )
      fprintf( stderr, "send_hand_soc: error: it!=%d but = %d\n", (int)size, it );

    if( nl )
      line = nl + 1;
    else
      line = NULL;
  }

  return( 1 );
}

/*******************************************************************************/
/* returns 0 if not, 1 if will, -1 otherwise    */
int net_connection_will_block( int sockfd, int on_send ) {
  fd_set fdset;
  struct timeval tv = { 0, 0 };
  int retry = 1, numfds = 0;
  NetRecvBuf *rb;

  wsa_get_ready();

  FD_ZERO( &fdset );
  FD_SET( sockfd, &fdset );

  while( retry ) {
    retry = 0;

    switch( on_send ) {
    case 0:
      numfds = select( sockfd + 1, &fdset, NULL, NULL, &tv );
      break;
    case 1:
      numfds = select( sockfd + 1, NULL, &fdset, NULL, &tv );
      break;
    case 2:
      numfds = select( sockfd + 1, &fdset, NULL, &fdset, &tv );
      break;
    }
    if( numfds == SOCKET_ERROR ) {
#if (PLATFORM_KIND!=Posix_Platform)
      errno = WSAGetLastError();
#endif
      if( errno == WSAEINTR ) {
	retry = 1;
      } else {
	perror( "net_connection_will_block: select" );
	return( -1 );
      }
    }
  }

  return( !numfds );
}

/*******************************************************************************/
/* returns bytes recieved if successful, -1 otherwise    */
int net_connection_recv( int sockfd, char *buf, uint len, int maxiter ) {
  int bytes, bytes_recv = 0;
  struct timeval tv = { 0, 10 };
  fd_set fdset;
  int retry = 1, i = 0;

  //--len;

  wsa_get_ready();

  while( retry ) {
    retry = 0;
    FD_ZERO( &fdset );
    FD_SET( sockfd, &fdset );

    if( select( sockfd + 1, &fdset, NULL, NULL, &tv ) == SOCKET_ERROR ) {
#if (PLATFORM_KIND!=Posix_Platform)
      errno = WSAGetLastError();
#endif
      if( !errno || errno == WSAEINTR ) {
	retry = 1;
      } else {
	perror( "net_connection_recv: select" );
	return( -1 );
      }
    } else {
      i++;
      if( (bytes = recv( sockfd, buf + bytes_recv, len - bytes_recv, 0 )) < 0 ) {
#if (PLATFORM_KIND!=Posix_Platform)
	errno = WSAGetLastError();
#endif
	if( errno == WSAEAGAIN || errno == WSAEINTR ) {
	  retry = i != maxiter;
	} else {
	  char ebuf[64];
	  sprintf( ebuf, "net_connection_recv: recv %d", errno );
	  perror( ebuf );
	  return( -1 );
	}
      } else {
	bytes_recv += bytes;
	if( bytes_recv < len && i != maxiter )
	  retry = 1;
      }
    }
  }

  return( bytes_recv );
}

/*******************************************************************************/

/* returns bytes recieved if successful, -1 otherwise    */
int recv_soc( int sockfd, char *line, int len )
{
 int bytes = net_connection_recv( sockfd, line, len - 1, 1 );
 line[bytes] = '\0';
 // if( bytes > -1 ) printf( "WF_RECV'D_SOC: '%s'\n", line );
 return bytes;
}


#ifndef NetAsync
typedef enum _NetConnectionType {
  NET_CONN_NONE = 0,
  NET_CONN_FILE,
  NET_CONN_INET,
  NET_CONN_SERV,
  NET_CONN_FILE_SERV
} NetConnectionType;

typedef struct NetAsync NetAsync;
struct NetAsync {
  NetConnectionType type;

  char *name;
  unsigned int ip;
  unsigned short port;

  char *remote;
  unsigned short remote_port;
  int socket;
  int state; // -1 = error, 0 = unconnected, 1 = connected

  void (*cb)( char *buf, uint len, void *user_data );
  void (*conn)( NetAsync *child, void *user_data );
  void *user_data;
  char *inbuf;
  uint inbuf_len;
  char *outbuf;
  uint outbuf_len;
  uint use_signals;

  NetAsync *parent;
  NetAsync *children;
  NetAsync *next;
};
#endif

NetAsync *net_async_new_client( const char *host, unsigned short port,
				void (*cb)( char *buf, uint n, void *user_data ),
				void *user_data, uint inbuf_len ) {
  NetAsync *na = c_new0( NetAsync, 1 );

  na->type = NET_CONN_INET;
  na->name = strdup( host );
  na->port = port;
  na->cb = cb;
  na->user_data = user_data;
  na->inbuf_len = inbuf_len;
  na->inbuf = malloc( inbuf_len );

  return( na );
}


void net_async_make_async( NetAsync *na ) {
#if (PLATFORM_KIND==Posix_Platform)
  pid_t pid;
  const int on = 1;

  wsa_get_ready();

  if( na->socket <= 0 )
    return;

  pid = getpid();
  //if( fcntl( na->socket, F_SETOWN, pid ) )
  if( ioctlsocket( na->socket, FIOSETOWN, &pid ) )
    perror( "net_async_make_async: fcntl(SETOWN)" );

  if( ioctlsocket( na->socket, FIOASYNC, &on ) )
    perror( "net_async_make_async: ioctl(IOASYNC)" );
#endif
}


NetAsync *net_async_new_server( unsigned short port,
				void (*conn)( NetAsync *child, void *user_data ),
				void (*cb)( char *buf, uint n, void *user_data ),
				void *user_data, uint inbuf_len ) {
  NetAsync *na = c_new0( NetAsync, 1 );

  na->type = NET_CONN_SERV;
  na->port = port;
  na->cb = cb;
  na->conn = conn;
  na->user_data = user_data;
  na->inbuf_len = inbuf_len;
  na->socket = net_server_new( port );


  return( na );
}


#ifndef memspool_flag
#define memspool_flag
char *memspool( char *spool, uint *len, const void *data, const size_t n ) {
  int newlen, modlen;

  if( !spool )
    *len = 0;
  if( !*len )
    spool = NULL;
  newlen = *len + n;
  modlen = (newlen + (1 << 6)) >> 6;

  if( !*len || modlen > ((*len + (1 << 6)) >> 6) ) {
    if( spool )
      spool = realloc( spool, modlen << 6 );
    else
      spool = malloc( modlen << 6 );
  }

  memcpy( spool + *len, data, n );
  *len = newlen;

  return( spool );
}
#endif


void net_async_send( NetAsync *na, char *bytes, uint len ) {
  na->outbuf = memspool( na->outbuf, &na->outbuf_len, bytes, len );
}


void net_async_delete( NetAsync *na ) {
  NetAsync *child;

  if( na->parent ) {
    if( na == na->parent->children ) {
      na->parent->children = na->parent->children->next;
    } else {
      child = na->parent->children;
      while( child->next && child->next != na )
	child = child->next;
      if( child->next )
	child->next = child->next->next;
    }
  }
  if( na->socket )
    closesocket( na->socket );
  free( na->inbuf );
  if( na->outbuf )
    free( na->outbuf );
  if( na->name )
    free( na->name );
  if( na->remote )
    free( na->remote );
  free( na );
}


int net_async_iteration( NetAsync *na ) {
  int bytes_recv, conns = 0, rv;
  NetAsync *child;

  wsa_get_ready();

  switch( na->type ) {
  case NET_CONN_INET:
    if( na->parent && !na->socket ) {
      na->type = NET_CONN_NONE;
      return( 0 );
    }

    if( !na->socket ) {
      if( !(na->socket = net_connection_new( na->name, na->port )) ) {
	na->state = -1;
	return( 0 );
      }
      if( na->use_signals )
	net_async_make_async( na );
      na->state = 1;
    }

    if( na->outbuf_len ) {
      rv = net_connection_send( na->socket, na->outbuf, na->outbuf_len );
      if( rv < 0 ) {
	closesocket( na->socket );
	na->socket = 0;
	puts( "na: got send error" );
      }
      free( na->outbuf );
      na->outbuf = NULL;
      na->outbuf_len = 0;
    }

    while( !net_connection_will_block( na->socket, 0 ) ) {
      bytes_recv = net_connection_recv( na->socket, na->inbuf,
					na->inbuf_len, 1 );
      if( bytes_recv == 0 ) {
	na->type = NET_CONN_NONE;
	return( 0 );
      } else if( bytes_recv > 0 && na->cb ) {
	na->cb( na->inbuf, bytes_recv, na->user_data );
      }
    }
    break;
  case NET_CONN_SERV:
    if( na->socket ) {
      while( !net_connection_will_block( na->socket, 0 ) ) {
	printf( "got conn %d\n", ++conns );
	child = c_new0( NetAsync, 1 );
	child->type = NET_CONN_INET;
	child->cb = na->cb;
	child->user_data = na->user_data;
	child->inbuf_len = na->inbuf_len;
	child->inbuf = malloc( na->inbuf_len );
	child->parent = na;
	child->socket = net_server_accept_connection( na->socket, child->inbuf,
						      na->inbuf_len,
						      &child->port );
	if( child->socket > 0 ) {
	  child->name = strdup( child->inbuf );
	  net_async_make_async( child );

	  if( na->conn )
	    na->conn( child, na->user_data );
	  child->next = na->children;
	  na->children = child;
	} else {
	  free( child );
	}
      }
    }

    child = na->children;
    while( child ) {
      net_async_iteration( child );
      child = child->next;
    }
    break;
  default:
    break;
  }

  return( 1 );
}


void net_async_server_close( NetAsync *na ) {
  NetAsync *child;

  wsa_get_ready();

  child = na->children;
  while( child ) {
    close( child->socket );
    child = child->next;
  }

  closesocket( na->socket );
}


int net_async_block_for( NetAsync *na, unsigned long timeout ) {
  fd_set fdset;
  struct timeval *tv = NULL, tvs = { 0, timeout };
  int retry = 1, numfds = 0, highfd = 0;
  NetAsync *nap;

  wsa_get_ready();

  if( timeout )
    tv = &tvs;

  FD_ZERO( &fdset );
  nap = na;
  do {
    if( (nap->type == NET_CONN_INET || nap->type == NET_CONN_SERV) && nap->socket ) {
      FD_SET( nap->socket, &fdset );
      if( nap->socket > highfd )
	highfd = nap->socket;
    }

    if( nap->children ) {
      nap = nap->children;
    } else if( nap != na ) {
      if( nap->next )
	nap = nap->next;
      else if( nap->parent != na && nap->parent->next )
	nap = nap->parent->next;
      else
	nap = na;
    }
  } while( nap != na );

  while( retry ) {
    retry = 0;

    numfds = select( highfd + 1, &fdset, &fdset, &fdset, tv );
    if( numfds == SOCKET_ERROR ) {
#if (PLATFORM_KIND!=Posix_Platform)
      errno = WSAGetLastError();
#endif
      if( errno == WSAEINTR ) {
	retry = 1;
      } else {
	printf( "error is %d\n", errno );
	perror( "net_async_block_for: select" );
	return( -1 );
      }
    }
  }

  return( !numfds );
}
